batch = {}

local _oobpoint1 = { 10000, 10000 }
local _oobpoint2 = { 10001, 10001 }

-- This isn't the most efficient way batch the metalines from an overdraw perspective.
-- It could easily group lines together that are on opposite sides of the screen.
function batch.simple( w, h, point1s, point2s, numlines, width )
	local batches = {}
	local pixels = 0
	local oobs = 0

	for i = 1, #point1s, numlines do
		local p1s = {}
		local p2s = {}
		local xmin, xmax = 100000, 0
		local ymin, ymax = 100000, 0

		for j = 0, numlines - 1 do
			local p1 = point1s[i+j]
			local p2 = point2s[i+j]

			if p1 and p2 then
				p1s[j+1] = p1
				p2s[j+1] = p2

				-- print(p1[1], h - p1[2], p2[1], h - p2[2])

				xmin = math.min(xmin, p1[1], p2[1])
				xmax = math.max(xmax, p1[1], p2[1])
				ymin = math.min(ymin, h - p1[2], h - p2[2])
				ymax = math.max(ymax, h - p1[2], h - p2[2])
			else
				-- Make a line that won't be rendered.
				p1s[j+1] = _oobpoint1
				p2s[j+1] = _oobpoint2

				oobs = oobs + 1
			end
		end

		-- print('minmax', xmin, xmax, ymin, ymax)

		local x = xmin - width
		local y = ymin - width
		local pw = (xmax - xmin) + 2 * width
		local ph = (ymax - ymin) + 2 * width

		local aabb = { x, y, pw, ph }

		pixels = pixels + (pw * ph)

		local batch = {
			point1s = p1s,
			point2s = p2s,
			aabb = aabb,
			pixels = pw * ph
		}

		batches[#batches+1] = batch
	end

	return {
		batches = batches,
		pixels = pixels,
		oobs = oobs,
	}
end

-- BIH stand for Bounded Interval Hierarchy
function batch.BIH( w, h, point1s, point2s, numlines, width )
	local batches = {}
	local pixels = 0
	local oobs = 0

	local function cmphorz( lhs, rhs )
		return math.min(lhs[1], lhs[3]) < math.min(rhs[1], rhs[3])
	end

	local function cmpvert( lhs, rhs )
		return math.min(lhs[2], lhs[4]) < math.min(rhs[2], rhs[4])
	end

	local function split( lines, horz )
		if #lines <= numlines then
			local point1s, point2s = {}, {}
			local xmin, xmax = 100000, 0
			local ymin, ymax = 100000, 0

			for i = 1, numlines do
				local line = lines[i]

				if line then
					local point1 = { line[1], line[2] }
					local point2 = { line[3], line[4] }

					point1s[i] = point1
					point2s[i] = point2

					xmin = math.min(xmin, point1[1], point2[1])
					xmax = math.max(xmax, point1[1], point2[1])
					ymin = math.min(ymin, h - point1[2], h - point2[2])
					ymax = math.max(ymax, h - point1[2], h - point2[2])
				else
					-- Make a line that won't be rendered.
					point1s[i] = _oobpoint1
					point2s[i] = _oobpoint2

					oobs = oobs + 1
				end
			end

			local x = xmin - width
			local y = ymin - width
			local pw = (xmax - xmin) + 2 * width
			local ph = (ymax - ymin) + 2 * width

			local aabb = { x, y, pw, ph }

			pixels = pixels + (pw * ph)

			local batch = {
				point1s = point1s,
				point2s = point2s,
				aabb = aabb,
				pixels = pw * ph
			}

			batches[#batches+1] = batch
		else
			table.sort(lines, (horz) and cmphorz or cmpvert)

			local mid = math.floor(0.5 + (#lines * 0.5))

			-- Lua doesn't have a table.sort(func, i, j) which is a great shame because
			-- we have to create copies of the arrays here :^(
			local lesslines, morelines = {}, {}

			for i = 1, mid do
				lesslines[i] = lines[i]
			end

			for i = mid + 1, #lines do
				morelines[i - mid] = lines[i]
			end

			split(lesslines, not horz)
			split(morelines, not horz)
		end
	end

	local lines = {}

	for i = 1, #point1s do
		local point1 = point1s[i]
		local point2 = point2s[i]

		lines[i] = { point1[1], point1[2], point2[1], point2[2] }
	end

	-- local horz = math.random() >= 0.5
	local horz = true

	split(lines, horz)

	return {
		batches = batches,
		pixels = pixels,
		oobs = oobs,
	}
end

-- Tries to make batches with the smallest area.
function batch.minarea( w, h, point1s, point2s, numlines, width )
	local batches = {}
	local pixels = 0
	local oobs = 0
	
	local function genaabb( line )
		local xmin = math.min(line[1], line[3])
		local xmax = math.max(line[1], line[3])
		local ymin = math.min(line[2], line[4])
		local ymax = math.max(line[2], line[4])

		return {
			xmin = xmin,
			ymin = ymin,
			xmax = xmax,
			ymax = ymax,
		}
	end

	local function aabbarea( aabb )
		return (aabb.xmax - aabb.xmin) * (aabb.ymax - aabb.ymin)
	end

	local function jointarea( aabb1, aabb2 )
		local xmin = math.min(aabb1.xmin, aabb2.xmin)
		local xmax = math.max(aabb1.xmax, aabb2.xmax)
		local ymin = math.min(aabb1.ymin, aabb2.ymin)
		local ymax = math.max(aabb1.ymax, aabb2.ymax)

		return (xmax - xmin) * (ymax - ymin)
	end

	local function mergeaabb( aabb1, aabb2 )
		local xmin = math.min(aabb1.xmin, aabb2.xmin)
		local xmax = math.max(aabb1.xmax, aabb2.xmax)
		local ymin = math.min(aabb1.ymin, aabb2.ymin)
		local ymax = math.max(aabb1.ymax, aabb2.ymax)

		return {
			xmin = xmin,
			ymin = ymin,
			xmax = xmax,
			ymax = ymax,
		}
	end

	local lines = {}

	for i = 1, #point1s do
		local point1 = point1s[i]
		local point2 = point2s[i]

		local line = { point1[1], point1[2], point2[1], point2[2], aabb = nil }
		line.aabb = genaabb(line)

		lines[i] = line
	end

	-- This means we always pick the smallest area line to start a bundle with.
	table.sort(lines,
		function ( lhs, rhs )
			return aabbarea(lhs.aabb) < aabbarea(rhs.aabb)
		end)

	while #lines > 0 do
		-- local pick = math.random(1, #lines)
		local pick = 1
		local bundle = { lines[pick] }
		local aabb = bundle[1].aabb

		lines[pick] = lines[#lines]
		lines[#lines] = nil

		while #bundle < numlines and #lines > 0 do
			local minarea = math.huge
			local lineindex = nil

			for i, line in ipairs(lines) do
				local area = jointarea(aabb, line.aabb)

				if area < minarea then
					minarea = area
					lineindex = i
				end
			end

			local line = lines[lineindex]
			bundle[#bundle+1] = line
			aabb = mergeaabb(aabb, line.aabb)

			lines[lineindex] = lines[#lines]
			lines[#lines] = nil
		end

		-- turn the bundle into a batch
		local point1s = {}
		local point2s = {}
		local xmin, xmax = 100000, 0
		local ymin, ymax = 100000, 0

		for i = 1, numlines do
			local line = bundle[i]

			if line then
				local point1 = { line[1], line[2] }
				local point2 = { line[3], line[4] }

				point1s[i] = point1
				point2s[i] = point2

				xmin = math.min(xmin, point1[1], point2[1])
				xmax = math.max(xmax, point1[1], point2[1])
				ymin = math.min(ymin, h - point1[2], h - point2[2])
				ymax = math.max(ymax, h - point1[2], h - point2[2])
			else
				-- Make a line that won't be rendered.
				point1s[i] = _oobpoint1
				point2s[i] = _oobpoint2

				oobs = oobs + 1
			end
		end

		local x = xmin - width
		local y = ymin - width
		local pw = (xmax - xmin) + 2 * width
		local ph = (ymax - ymin) + 2 * width

		local aabb = { x, y, pw, ph }

		pixels = pixels + (pw * ph)

		local batch = {
			point1s = point1s,
			point2s = point2s,
			aabb = aabb,
			pixels = pw * ph
		}

		batches[#batches+1] = batch
	end

	return {
		batches = batches,
		pixels = pixels,
		oobs = oobs,
	}
end



function table.pick( tbl, index )
	local result = tbl[index]
	
	tbl[index] = tbl[#tbl]
	tbl[#tbl] = nil

	return result
end

--[[
-- limits = {
--     minwidth = minimum width of a resulting bbox
--     minheight = minimum height of a resulting bbox
--     margin = minimum space between resulting bboxes
--     maxleaves = maximum no. of resulting leaves.
-- }
--

local _axes = {
	'horz',
	'vert',
}

function partitiongen.bsp( bbox, limits )
	local minbranchwidth = (2 * limits.minwidth) + (2 * limits.margin)
	local minbranchheight = (2 * limits.minheight) + (2 * limits.margin)

	local minleafwidth = limits.minwidth + limits.margin
	local minleafheight = limits.minheight + limits.margin

	-- A branch is an aabb that is big enough to be split.
	local branches = { bbox:clone() }
	-- A leaf is an aabb that can't be split without making it smaller than
	-- the limits allow.
	local leaves = {}

	while #branches > 0 or #branches + #leaves < limits.maxleaves do
		local branch = table.pick(branches, math.random(1, #branches))

		local horzsplit = branch:width() > minbranchwidth
		local vertsplit = branch:height() > minbranchheight

		if not horzsplit and not vertsplit then
			leaves[#leaves+1] = branch
		else
			local axis = _axes[math.random(horzsplit and 1 or 2, vertsplit and 2 or 1)]
			local coord = nil

			if axis == 'horz' then
				local min = branch.xmin + minleafwidth
				local max = branch.xmax - minleafwidth
				coord = math.random(min, max)
			else
				local min = branch.ymin + minleafheight
				local max = branch.ymax - minleafheight
				coord = math.random(min, max)
			end

			local sub1, sub2 = branch:split(axis, coord)

			if sub1:width() < minbranchwidth and sub1:height() < minbranchheight then
				leaves[#leaves+1] = sub1
			else
				branches[#branches+1] = sub1
			end

			if sub2:width() < minbranchwidth and sub2:height() < minbranchheight then
				leaves[#leaves+1] = sub2
			else
				branches[#branches+1] = sub2
			end
		end
	end

	local result = {}

	for _, bbox in ipairs(open) do
		result[#result+1] = bbox.shrink(limits.margin * 0.5)
	end

	for _, bbox in ipairs(closed) do
		result[#result+1] = bbox.shrink(limits.margin * 0.5)
	end

	return result
end
--]]